@@ -60,24 +60,29 @@ class DialogueFlow:
         return _autostate
 
     def __init__(self, initial_state: Union[Enum, str, tuple], initial_speaker = Speaker.SYSTEM,
-                 macros: Dict[str, Macro] =None, kb: Union[KnowledgeBase, str, List[str]] =None, default_system_state=None):
+                 macros: Dict[str, Macro] =None, kb: Union[KnowledgeBase, str, List[str]] =None,
+                 default_system_state=None, end_state='end', all_multi_hop=True, wordnet=False):
         self._graph = GraphDatabase()
         self._initial_state = State(initial_state)
         self._potential_transition = None
         self._initial_speaker = initial_speaker
         self._speaker = self._initial_speaker
-        self._response = None
         self._vars = HashableDict()
+        self._transitions = []
+        self.vars()['__state__'] = self._initial_state
         self.set_state(self._initial_state)
-        self._gate_requirements = defaultdict(dict)
-        self._gates = defaultdict(set)
-        self._gate_buffer = {}
+        self._gates = defaultdict(list)
         self._prepends = {}
         self._var_dependencies = defaultdict(set)
         self._error_transitioned = False
-        self._is_module = False
         self._default_state = default_system_state
+        self._end_state = end_state
+        self._goals = {}
+        self._all_multi_hop = all_multi_hop
+        self._composite_dialogue_flow = None
+        self._namespace = None
         self.vars()['__stack__'] = []
+        self.vars()['__system_state__'] = 'None' if initial_speaker == Speaker.USER else self._initial_state
         if kb is None:
             self._kb = KnowledgeBase()
         elif isinstance(kb, str):
@@ -91,8 +96,10 @@ class DialogueFlow:
             self._kb = kb
         onte = ONTE(self._kb)
         kbe = KBE(self._kb)
+        goal_exit_macro = GoalExit(self)
+        conjunction_macro = CheckVarsConjunction()
         self._macros = {
-            'WN': WN(),
+            'WN': WN(wordnet),
             'ONT': onte, 'ONTE': onte,
             'KBQ': kbe, 'KBE': kbe,
             'ONTN': ONTN(self._kb),
@@ -103,7 +110,8 @@ class DialogueFlow:
             'I': Intersection(),
             'DIF': Difference(),
             'SET': SetVars(),
-            'ALL': CheckVarsConjunction(),
+            'ALL': conjunction_macro,
+            'IF': conjunction_macro,
             'ANY': CheckVarsDisjunction(),
             'ISP': IsPlural(),
             'FPP': FirstPersonPronoun(self._kb),
@@ -111,6 +119,7 @@ class DialogueFlow:
             'PSP': PossessivePronoun(self._kb),
             'EQ': Equal(),
             'GATE': Gate(self),
+            'UNSET': Unset(),
             'CLR': Clear(),
             'NER': NamedEntity(),
             'POS': PartOfSpeech(),
@@ -123,12 +132,24 @@ class DialogueFlow:
             'NEGATION': Negation(),
             'IDK': DontKnow(),
             'MAYBE': Maybe(),
-            'TRANSITION': Transition(),
-            'UNX': Unexpected()
+            'CONFIRM': Confirm(),
+            'TRANSITION': Transition(self),
+            'UNX': Unexpected(),
+            'INT': Intent(),
+            'GEXT': goal_exit_macro,
+            'GOAL': GoalPursuit(goal_exit_macro),
+            'GCOM': GoalCompletion(self),
+            'GRET': GoalReturn(self),
+            'DEFAULT': Default(),
+            'VT': VirtualTransitions(self),
+            'GSRET': SetGoalReturnPoint(),
+            'TARGET': Target()
         }
         if macros:
             self._macros.update(macros)
         self._rules = UpdateRules(vars=self._vars, macros=self._macros)
+        self.add_state('end')
+
 
     # TOP LEVEL: SYSTEM-LEVEL USE CASES
 
@@ -138,7 +159,7 @@ class DialogueFlow:
         :return: None
         """
         t1 = time()
-        while True:
+        while self.state() != self.end_state():
             if self.speaker() == Speaker.SYSTEM:
                 system_response = self.system_turn(debugging=debugging)
                 if debugging:
@@ -156,15 +177,15 @@ class DialogueFlow:
         :return: the natural language system response
         """
         t1 = time()
+        self.vars()['__goal_return_state__'] = 'None'
         visited = {self.state()}
         responses = []
         while self.speaker() is Speaker.SYSTEM:
             response, next_state = self.system_transition(self.state(), debugging=debugging)
-            self.take_transition(next_state)
+            self.set_state(next_state)
             responses.append(response)
-            if next_state in visited and self._speaker is Speaker.SYSTEM:
-                self.change_speaker()
-                break
+            if next_state in visited or (not self.state_settings(next_state).system_multi_hop):
+                self.set_speaker(Speaker.USER)
             visited.add(next_state)
         t2 = time()
         if debugging:
@@ -180,7 +201,8 @@ class DialogueFlow:
         :return: None
         """
         t1 = time()
-        self.state_update(natural_language, debugging)
+        self._transitions.clear()
+        self.apply_update_rules(natural_language, debugging)
         visited = {self.state()}
         while self.speaker() is Speaker.USER:
             next_state = self.user_transition(natural_language, self.state(), debugging=debugging)
@@ -192,10 +214,9 @@ class DialogueFlow:
                 except RuntimeError:
                     if debugging:
                         print("Couldn't error hop")
-            self.take_transition(next_state)
-            if next_state in visited and self._speaker is Speaker.USER:
-                self.change_speaker()
-                break
+            self.set_state(next_state)
+            if next_state in visited or (not self.state_settings(next_state).user_multi_hop):
+                self.set_speaker(Speaker.SYSTEM)
             visited.add(next_state)
         self.set_speaker(Speaker.SYSTEM)
         t2 = time()
@@ -203,7 +224,7 @@ class DialogueFlow:
             print('User turn in {:.5f}'.format(t2 - t1))
 
 
-    def load_transitions(self, json_dict, speaker=None, root=None):
+    def load_transitions(self, json_dict, speaker=None):
         """
         wheeeeeeee!
         """
@@ -213,11 +234,10 @@ class DialogueFlow:
             source = json_dict['state']
         else:
             source = DialogueFlow.autostate()
-        if root is None:
-            root = self._initial_state
 
         hop = None
         switch = False
+        enter = None
 
         # read settings and transitions for state
         transitions = []
@@ -234,13 +254,14 @@ class DialogueFlow:
                 self.set_state_prepend(source, prepend)
             elif key == 'switch':
                 switch = json_dict['switch']
-            elif key not in {'state', 'hop', 'score', 'switch'}:
+            elif key == 'enter':
+                enter = json_dict['enter']
+            elif key not in {'state', 'hop', 'score', 'switch', 'enter'}:
                 transitions.append((key, value))
 
         # set up state settings
         if not self.has_state(source):
             self.add_state(source)
-            self.set_error_successor(source, root)
         if hop:
             if speaker == Speaker.USER:
                 speaker = Speaker.SYSTEM
@@ -250,6 +271,8 @@ class DialogueFlow:
                 self.state_settings(source).update(user_multi_hop=True)
         if switch:
             self.update_state_settings(source, switch=True)
+        if enter:
+            self.update_state_settings(source, enter=enter)
 
         # set up transitions
         expanded_transitions = []
@@ -269,7 +292,6 @@ class DialogueFlow:
                     target = target['state']
                     if not self.has_state(target):
                         self.add_state(target)
-                        self.set_error_successor(target, root)
                 self.set_error_successor(source, target)
 
             else:
@@ -283,20 +305,18 @@ class DialogueFlow:
                     target = target['state']
                     if not self.has_state(target):
                         self.add_state(target)
-                        self.set_error_successor(target, root)
                 if speaker == Speaker.USER:
                     if self.has_transition(source, target, Speaker.USER):
-                        intermediate = self.autostate()
-                        self.add_state(intermediate, target, user_multi_hop=True)
-                        self.add_user_transition(source, intermediate, natex, score=score)
+                        intermediate = '_' + self.autostate()
+                        self.add_state(intermediate, target)
+                        self.add_user_transition(source, intermediate, natex + ' #TARGET(%s)' % target, score=score)
                     else:
                         self.add_user_transition(source, target, natex, score=score)
                 elif speaker == Speaker.SYSTEM:
                     if self.has_transition(source, target, Speaker.SYSTEM):
-                        intermediate = self.autostate()
-                        self.add_state(intermediate, system_multi_hop=True)
-                        self.add_system_transition(intermediate, target, '')
-                        self.add_system_transition(source, intermediate, natex, score=score)
+                        intermediate = '_' + self.autostate()
+                        self.add_state(intermediate)
+                        self.add_system_transition(source, intermediate, natex + ' #TARGET(%s)' % target, score=score)
                     else:
                         self.add_system_transition(source, target, natex, score=score)
 
@@ -324,23 +344,20 @@ class DialogueFlow:
             state = self.state()
         else:
             state = State(state)
-        transition_options = {}
-        if self._response is not None:
-            response, vars, dest, score = self._response
-            transition_options[(response, (self.state(), dest, self._speaker), vars)] = score
-        self._gate_buffer.clear()
+        transition_options = []
         transitions = list(self.transitions(state, Speaker.SYSTEM))
+        transition_items = []
         for transition in transitions:
-            self._potential_transition = transition
-            if not self._is_module and isinstance(transition[1], tuple):
-                continue
-            t1 = time()
             natex = self.transition_natex(*transition)
-            if '->' in transition[1]:
-                transition = (transition[1].split('->')[0], transition[1].split('->')[1], Speaker.SYSTEM)
-                natex = natex + self.transition_natex(*transition)
-            settings = self.transition_settings(*transition)
+            score = self.transition_settings(*transition).score
+            transition_items.append((natex, transition, score))
+        while self._transitions:
+            natex, transition, score = self._transitions.pop()
+            transition_items.append((natex, transition, score))
+        for natex, transition, score in transition_items:
+            t1 = time()
             vars = HashableDict(self._vars)
+            self._potential_transition = transition # MOVED, todo
             try:
                 generation = natex.generate(vars=vars, macros=self._macros, debugging=debugging)
             except Exception as e:
@@ -349,42 +366,89 @@ class DialogueFlow:
                 print('Transition {}: {} failed'.format(str(transition), natex))
                 print()
                 generation = None
-            if generation is not None:
+            source, target, speaker = transition
+            if '__source__' in vars:
+                source = State(module_state(vars['__source__']))
+                del vars['__source__']
+            if '__target__' in vars:
+                target = State(module_state(vars['__target__']))
+                del vars['__target__']
+            transition = source, target, speaker
+            if not self.is_module() and isinstance(target, tuple):
+                continue
+            if '->' in transition[1]:
+                transition = (target.split('->')[0], target.split('->')[1], speaker)
+                try:
+                    appended_generation = self.transition_natex(*transition).generate(vars=vars, macros=self._macros, debugging=debugging)
+                    generation = generation + ' ' + appended_generation
+                except Exception as e:
+                    print()
+                    print(e)
+                    print('Transition {}: {} failed'.format(str(transition), natex))
+                    print()
+                    generation = None
+            elif isinstance(transition[1], tuple) and '->' in transition[1][1]:
+                namespace = transition[1][0]
+                transition = ((namespace, target[1].split('->')[0]), (namespace, target[1].split('->')[1]), speaker)
+                try:
+                    appended_generation = self.composite_dialogue_flow().transition_natex(
+                        namespace, *transition).generate(vars=vars, macros=self._macros, debugging=debugging)
+                    generation = generation + ' ' + appended_generation
+                except Exception as e:
+                    print()
+                    print(e)
+                    print('Transition {}: {} failed'.format(str(transition), natex))
+                    print()
+                    generation = None
+            source, target, speaker = transition
+            if '__source__' in vars:
+                source = State(module_state(vars['__source__']))
+                del vars['__source__']
+            if '__target__' in vars:
+                target = State(module_state(vars['__target__']))
+                del vars['__target__']
+            transition = source, target, speaker
+            if self.is_module() and isinstance(target, tuple):
+                enter_natex = self.composite_dialogue_flow().state_settings(*target).enter
+            else:
+                enter_natex = self.state_settings(target).enter
+            enter_natex_pass = True
+            if enter_natex is not None:
+                try:
+                    enter_natex_pass = enter_natex.generate(vars=vars, macros=self._macros, debugging=debugging)
+                except Exception as e:
+                    print()
+                    print(e)
+                    print('Enter Natex {}: {} failed'.format(str(target), enter_natex))
+                    print()
+                    enter_natex_pass = None
+            if generation is not None and enter_natex_pass is not None:
                 if '__score__' in vars:
                     score = vars['__score__']
                     del vars['__score__']
-                else:
-                    score = settings.score
-                transition_options[(generation, transition, vars)] = score
+                gate_closed = False
+                gate_var_config = None
+                gate_target_id = None
+                if '__gate__' in vars:
+                    gate_var_config = vars['__gate__']
+                    gate_target_id = (self.namespace(), target) if (not isinstance(target, tuple) and self.is_module()) else target
+                    for vc in self.gates()[gate_target_id]:
+                        if gate_var_config == vc:
+                            gate_closed = True
+                    del vars['__gate__']
+                if not gate_closed:
+                    transition_options.append((score, generation, transition, vars, gate_var_config, gate_target_id))
             t2 = time()
             if debugging:
                 print('Transition {} evaluated in {:.5f}'.format(transition, t2-t1))
+            while self._transitions:
+                natex, transition, score = self._transitions.pop()
+                transition_items.append((natex, transition, score))
+        self._transitions.clear()
         if transition_options:
-            remaining = list(transition_options)
-            memory = self.state_settings(state).memory
-            for item in memory:
-                if len(transition_options) > 1:
-                    key = None
-                    for k in transition_options:
-                        if k[1] == item:
-                            key = k
-                            break
-                    if key:
-                        del transition_options[key]
-                else:
-                    break
-            response, transition, vars = random_max(transition_options, key=lambda x: transition_options[x])
-            if self.is_switch(state):
-                to_add_to_stack = [x[1] for x in remaining if x != (response, transition, vars)]
-                for trans in to_add_to_stack:
-                    src, tgt, natex = trans[0], trans[1], self.transition_natex(*trans)
-                    if isinstance(src, tuple):
-                        src = ':'.join(src)
-                    if isinstance(tgt, tuple):
-                        tgt = ':'.join(tgt)
-                    if src != state:
-                        tgt = src + '->' + tgt
-                    self.vars()['__stack__'].append(tgt)
+            score, response, transition, vars, gate_var_config, gate_target_id = random_max(transition_options, key=lambda x: x[0])
+            if gate_var_config is not None:
+                self.gates()[gate_target_id].append(gate_var_config)
             if debugging:
                 updates = {}
                 for k, v in vars.items():
@@ -397,63 +461,23 @@ class DialogueFlow:
                             print('  {} = {} -> {}'.format(k, self._vars[k], v))
                         else:
                             print('  {} = None -> {}'.format(k, v))
-            if transition in self.gate_buffer():
-                self.gates()[transition].add(self.gate_buffer()[transition])
             self.update_vars(vars)
             next_state = transition[1]
             if debugging:
                 tf = time()
                 print('System transition in {:.5f}'.format(tf-ti))
                 print('Transitioning {} -> {}'.format(self.state(), next_state))
-            if self._response is not None:
-                self._response = None
             if '__response_prefix__' in self.vars() and self.vars()['__response_prefix__'] != 'None':
                 response = self.vars()['__response_prefix__'] + ' ' + response
                 self.vars()['__response_prefix__'] = 'None'
-            if self.is_switch(state):
-                switch_response, switch_next_state = self.system_transition(next_state)
-                return response + ' ' + switch_response, switch_next_state
             return response, next_state
         else:
-            while self.is_switch(state) and self.vars()['__stack__']:
-                target = self.vars()['__stack__'].pop()
-                if '->' in target:
-                    source, target = target.split('->')
-                    transition = source, target, Speaker.SYSTEM
-                    natex = self.transition_natex(*transition)
-                    vars = HashableDict(self._vars)
-                    try:
-                        response = natex.generate(vars=vars, macros=self._macros, debugging=debugging)
-                    except Exception as e:
-                        print()
-                        print(e)
-                        print('Stack transition to {}: {} failed'.format(str(target), natex))
-                        print()
-                        response = None
-                    if response is not None:
-                        if transition in self.gate_buffer():
-                            self.gates()[transition].add(self.gate_buffer()[transition])
-                        self.update_vars(vars)
-                        next_state = transition[1]
-                        if debugging:
-                            tf = time()
-                            print('System transition in {:.5f}'.format(tf - ti))
-                            print('Transitioning {} -> {}'.format(self.state(), next_state))
-                        if self._response is not None:
-                            self._response = None
-                        if '__response_prefix__' in self.vars() and self.vars()['__response_prefix__'] != 'None':
-                            response = self.vars()['__response_prefix__'] + ' ' + response
-                            self.vars()['__response_prefix__'] = 'None'
-                        return response, next_state
-                else:
-                    return self.system_transition(target, debugging=debugging)
-            else:
-                if self._default_state is not None:
-                    self.set_state(self._default_state)
-                    if debugging:
-                        print('No valid system transitions found, going to default state...')
-                    return self.system_transition(self.state(), debugging=debugging)
-                raise AssertionError('dialogue flow system transition found no valid options')
+            if self._default_state is not None:
+                self.set_state(self._default_state)
+                if debugging:
+                    print('No valid system transitions found, going to default state...')
+                return self.system_transition(self.state(), debugging=debugging)
+            raise AssertionError('dialogue flow system transition found no valid options')
 
 
     def user_transition(self, natural_language: str, state: Union[Enum, str, tuple], debugging=False):
@@ -464,6 +488,7 @@ class DialogueFlow:
         :return: the successor state representing the highest score user transition
                  that matches natural_language, or None if none match
         """
+        natural_language = ''.join([c.lower() for c in natural_language if c.isalpha() or c == ' '])
         state = module_state(state)
         self._error_transitioned = False
         ti = time()
@@ -472,22 +497,22 @@ class DialogueFlow:
         else:
             state = State(state)
         transition_options = []
-        if '__transition__' in self.vars() and self.vars()['__transition__']:
-            transition_options.append(
-                (self.vars()['__transition_score__'],
-                 ('__global__', self.vars()['__transition__'], Speaker.USER),
-                 self.vars()))
-        ngrams = Ngrams(natural_language, n=10)
-        self._gate_buffer.clear()
+        transition_items = []
         for transition in self.transitions(state, Speaker.USER):
+            natex = self.transition_natex(*transition)
+            score = self.transition_settings(*transition).score
+            transition_items.append((natex, transition, score))
+        while self._transitions:
+            natex, transition, score = self._transitions.pop()
+            transition_items.append((natex, transition, score))
+        ngrams = Ngrams(natural_language, n=10)
+        for natex, transition, score in transition_items:
             self._potential_transition = transition
-            if not self._is_module and isinstance(transition[1], tuple):
+            if not self.is_module() and isinstance(transition[1], tuple):
                 continue
             t1 = time()
             if debugging:
                 print('Evaluating transition {}'.format(transition[:2]))
-            natex = self.transition_natex(*transition)
-            settings = self.transition_settings(*transition)
             vars = HashableDict(self._vars)
             try:
                 match = natex.match(natural_language, vars, self._macros, ngrams, debugging)
@@ -497,20 +522,58 @@ class DialogueFlow:
                 print('Transition {}: {} failed'.format(str(transition), natex))
                 print()
                 match = None
-            if match:
+            source, target, speaker = transition
+            if '__source__' in vars:
+                source = State(module_state(vars['__source__']))
+                del vars['__source__']
+            if '__target__' in vars:
+                target = State(module_state(vars['__target__']))
+                del vars['__target__']
+            transition = source, target, speaker
+            if self.is_module() and isinstance(target, tuple):
+                enter_natex = self.composite_dialogue_flow().state_settings(*target).enter
+            else:
+                enter_natex = self.state_settings(target).enter
+            enter_natex_pass = True
+            if enter_natex is not None:
+                try:
+                    enter_natex_pass = enter_natex.generate(vars=vars, macros=self._macros, debugging=debugging)
+                except Exception as e:
+                    print()
+                    print(e)
+                    print('Enter Natex {}: {} failed'.format(str(target), enter_natex))
+                    print()
+                    enter_natex_pass = None
+            if match and enter_natex_pass is not None:
                 if debugging:
                     print('Transition {} matched "{}"'.format(transition[:2], natural_language))
                 if '__score__' in vars:
                     score = vars['__score__']
                     del vars['__score__']
-                else:
-                    score = settings.score
-                transition_options.append((score, transition, vars))
+                gate_closed = False
+                gate_var_config = None
+                gate_target_id = None
+                if '__gate__' in vars:
+                    gate_var_config = vars['__gate__']
+                    gate_target_id = (self.namespace(), target) if (
+                                not isinstance(target, tuple) and self.is_module()) else target
+                    for vc in self.gates()[gate_target_id]:
+                        if gate_var_config == vc:
+                            gate_closed = True
+                    del vars['__gate__']
+                if not gate_closed:
+                    transition_options.append((score, transition, vars, gate_var_config, gate_target_id))
             t2 = time()
             if debugging:
                 print('Transition {} evaluated in {:.5f}'.format(transition, t2-t1))
+            while self._transitions:
+                natex, transition, score = self._transitions.pop()
+                transition_items.append((natex, transition, score))
+        self._transitions.clear()
         if transition_options:
-            score, transition, vars = random_max(transition_options, key=lambda x: x[0])
+            score, transition, vars, gate_var_config, gate_target_id = random_max(transition_options, key=lambda x: x[0])
+            if gate_var_config is not None:
+                self.gates()[gate_target_id].append(gate_var_config)
             if debugging:
                 updates = {}
                 for k, v in vars.items():
@@ -523,8 +586,6 @@ class DialogueFlow:
                             print('  {} = {} -> {}'.format(k, self._vars[k], v))
                         else:
                             print('  {} = None -> {}'.format(k, v))
-            if transition in self.gate_buffer():
-                self.gates()[transition].add(self.gate_buffer()[transition])
             self.update_vars(vars)
             next_state = transition[1]
             if debugging:
@@ -612,6 +673,8 @@ class DialogueFlow:
         self.set_transition_natex(source, target, Speaker.USER, natex_nlu)
         transition_settings = Settings(score=1.0)
         transition_settings.update(**settings)
+        if self._all_multi_hop:
+            self.update_state_settings(source, user_multi_hop=True)
         self.set_transition_settings(source, target, Speaker.USER, transition_settings)
         if target in self._prepends:
             prepend = self._prepends[target]
@@ -635,6 +698,8 @@ class DialogueFlow:
         transition_settings = Settings(score=1.0)
         transition_settings.update(**settings)
         self.set_transition_settings(source, target, Speaker.SYSTEM, transition_settings)
+        if self._all_multi_hop:
+            self.update_state_settings(source, system_multi_hop=True)
         if target in self._prepends:
             prepend = self._prepends[target]
             natex = self.transition_natex(source, target, Speaker.SYSTEM)
@@ -645,7 +710,7 @@ class DialogueFlow:
         state = State(state)
         if self.has_state(state):
             raise ValueError('state {} already exists'.format(state))
-        state_settings = Settings(user_multi_hop=False, system_multi_hop=False, switch=False, memory=10)
+        state_settings = Settings(user_multi_hop=False, system_multi_hop=False, switch=False, enter=None)
         state_settings.update(**settings)
         self._graph.add_node(state)
         self.update_state_settings(state, **state_settings)
@@ -654,21 +719,6 @@ class DialogueFlow:
             self.set_error_successor(state, error_successor)
 
 
-    # MID LEVEL
-
-    def take_transition(self, target):
-        target = module_state(target)
-        if self.speaker() is Speaker.SYSTEM:
-            transition = (self.state(), target, self.speaker())
-            self.state_settings(self.state()).memory.add(transition)
-        self.set_state(target)
-        if self.speaker() is Speaker.SYSTEM:
-            if not self.state_settings(self.state()).system_multi_hop:
-                self.set_speaker(Speaker.USER)
-        else:
-            if not self.state_settings(self.state()).user_multi_hop:
-                self.set_speaker(Speaker.SYSTEM)
-
     # LOW LEVEL: PROPERTIES, GETTERS, SETTERS
 
     def transition_natex(self, source: Union[Enum, str, tuple], target: Union[Enum, str, tuple], speaker: Enum):
@@ -722,10 +772,10 @@ class DialogueFlow:
         state = State(state)
         if 'settings' not in self._graph.data(state):
             self._graph.data(state)['settings'] = Settings()
-        if 'memory' in settings:
-            settings['memory'] = Memory(settings['memory'])
         if 'global_nlu' in settings:
             self.add_global_nlu(state, settings['global_nlu'])
+        if 'enter' in settings and isinstance(settings['enter'], str):
+            settings['enter'] = NatexNLG(settings['enter'], macros=self._macros)
         self.state_settings(state).update(**settings)
 
     def remove_transition(self, source, target, speaker):
@@ -743,6 +793,15 @@ class DialogueFlow:
     def set_state(self, state: Union[Enum, str, tuple]):
         state = module_state(state)
         state = State(state)
+        if self.speaker() == Speaker.SYSTEM:
+            if '__state__' in self.vars():
+                st_str = self.vars()['__state__'][1] if isinstance(self.vars()['__state__'],tuple) else self.vars()['__state__']
+                if not st_str.startswith('_'):
+                    self.vars()['__system_state__'] = self.vars()['__state__']
+                if '__system_state__' not in self.vars():
+                    self.vars()['__system_state__'] = 'None'
+            else:
+                self.vars()['__system_state__'] = 'None'
         self._vars['__state__'] = state
 
     def has_state(self, state):
@@ -816,24 +875,26 @@ class DialogueFlow:
             self.set_speaker(Speaker.USER)
 
     def reset(self):
+        self._transitions.clear()
         self._speaker = self._initial_speaker
         self._vars = HashableDict()
+        self.vars()['__state__'] = self._initial_state
+        self.vars()['__stack__'] = []
+        self.vars()['__system_state__'] = 'None' if self._initial_speaker == Speaker.USER else self._initial_state
         self.set_state(self._initial_state)
         self._rules.set_vars(self._vars)
         self._gates = defaultdict(set)
-        for state in self.graph().nodes():
-            self.state_settings(state).memory.clear()
 
     def update_vars(self, variables: HashableDict):
         if not isinstance(variables, HashableDict):
             variables = HashableDict(variables)
-        for k in variables.altered():
+        for k in variables:
             if k in self._var_dependencies:
                 dependencies = self._var_dependencies[k]
                 for dependency in dependencies:
                     if dependency in self._vars:
                         self._vars[dependency] = None
-        self._vars.update({k: variables[k] for k in variables.altered()})
+        self._vars.update({k: variables[k] for k in variables if k != '__score__' and k in variables})
 
     def potential_transition(self):
         return self._potential_transition
@@ -841,18 +902,6 @@ class DialogueFlow:
     def gates(self):
         return self._gates
 
-    def gate_buffer(self):
-        return self._gate_buffer
-
-    def buffer_configuration(self, configuration):
-        self._gate_buffer[self._potential_transition] = configuration
-
-    def gate_requirements(self):
-        return self._gate_requirements
-
-    def set_gate_requirements(self, requirements):
-        self._gate_requirements[self._potential_transition] = requirements
-
     def var_dependencies(self):
         return self._var_dependencies
 
@@ -864,41 +913,80 @@ class DialogueFlow:
                 natex = self.transition_natex(*transition)
                 self.set_transition_natex(*transition, prepend + natex)
 
-    def passes_gate(self, var_config):
-        if var_config in self._gates[self._potential_transition]:
-            return False
-        for k, v in self._gate_requirements[self._potential_transition].items():
-            if k not in var_config:
-                if v is not None:
-                    return False
-            else:
-                if v != var_config[k]:
-                    return False
-        for k, v in var_config.items():
-            if v is None:
-                if k not in self._gate_requirements[self._potential_transition] \
-                    or self._gate_requirements[self._potential_transition][k] is not None:
-                    return False
-        return True
-
     def add_update_rule(self, precondition, postcondition=None):
         self._rules.add(precondition, postcondition)
 
-    def state_update(self, user_input, debugging=False):
+    def apply_update_rules(self, user_input, debugging=False):
         result = self._rules.update(user_input, debugging)
         if result is not None:
             response, score = result
-            self._response = response, self._vars, self.state(), score
+            self._transitions.append(
+                (response, (self.state(), self.state(), Speaker.SYSTEM), score))
             self.set_speaker(Speaker.SYSTEM)
 
     def knowledge_base(self):
         return self._kb
 
-    def set_is_module(self):
-        self._is_module = True
+    def set_is_module(self, composite_dialogue_flow):
+        self._composite_dialogue_flow = composite_dialogue_flow
 
     def is_switch(self, state):
         return self.state_settings(state)['switch']
 
+    def end_state(self):
+        return self._end_state
+
     def update_rules(self):
-        return self._rules
\ No newline at end of file
+        return self._rules
+
+    def goals(self):
+        return self._goals
+
+    def set_goals(self, goals_dict):
+        self._goals = goals_dict
+
+    def dynamic_transitions(self):
+        return self._transitions
+
+    def composite_dialogue_flow(self):
+        return self._composite_dialogue_flow
+
+    def is_module(self):
+        return self.composite_dialogue_flow() is not None
+
+    def namespace(self):
+        return self._namespace
+
+    def set_namespace(self, namespace):
+        self._namespace = namespace
+
+    def set_gates(self, gates):
+        self._gates = gates
+
+    def load_global_nlu(self, transitions):
+        for nlu, followup in transitions.items():
+            if nlu == 'state':
+                continue
+            score = 0.5
+            if isinstance(followup, str):
+                state = followup
+            else:
+                if 'state' not in followup:
+                    state = DialogueFlow.autostate()
+                    followup['state'] = state
+                else:
+                    state = followup['state']
+                if 'score' in followup:
+                    score = followup['score']
+            nlu = nlu + ' #GEXT'
+            self.add_global_nlu(state, nlu, score)
+        self.load_transitions(transitions, Speaker.USER)
+
+    def add_goal(self, id_string, return_state=None, return_phrase=None, doom_counter=None):
+        goal = {
+            'id': id_string,
+            'return_state': return_state,
+            'return_phrase': return_phrase,
+            'doom_counter': doom_counter
+        }
+        self._goals[id_string] = goal
